#include "GameDataStruct.h"
#include "dijkstra.h"

double DistPoints(double lx, double ly, double rx, double ry )
{
	double len = ( rx - lx )*( rx - lx ) + ( ry - ly )*( ry - ly );
	
	return pow( len , 0.5 );
}

//Oρ@Ɠ_̋oƂɎg
double CrossProduct(double lx, double ly, double rx, double ry ) { return (lx * ry) - (ly * rx); }

//ρ@xNg̃`FbNɂ
double DotProduct(double lx, double ly, double rx, double ry ) { return (lx * rx) + (ly * ry); }

//ɓ_邩
bool IntersectPointLine(
	int x, int y,		//_
	int lx1, int ly1,	//n_
	int lx2, int ly2,	//I_
	double limit )		//덷
{
	double dist;
	
	do {

		double lineLen = DistPoints(lx1, ly1, lx2, ly2);
		if ( lineLen == 0.0 ) {
			dist = DistPoints( x, y, lx1, ly1 );
			break;
		}
		
		if (DotProduct(lx2 - lx1, ly2 - ly1, x - lx1, y- ly1) <= 0) {
			dist = DistPoints( x, y, lx1, ly1 );
			break;
        }

		if (DotProduct(lx1 - lx2, ly1 - ly2, x - lx2, y- ly2) <= 0) {
			dist = DistPoints( x, y, lx2, ly2 );
            break;
        }
		
		dist = CrossProduct(lx2 - lx1, ly2 - ly1, x - lx1, y- ly1) / lineLen;
		dist = abs(dist);

	} while(0);
	
	if ( dist < limit) {
		return true;
	}
	return false;
}


//m[hԂ̃RXg()vZ
double QueryCost(Node* n1, Node* n2 )
{
	return DistPoints(n1->x,n1->y, n2->x, n2->y);
}

Dijkstra::Dijkstra()
{
	m_start = NULL;
	m_goal = NULL;
}
Dijkstra::~Dijkstra()
{
	Clear();
}

//wm[h폜
void Dijkstra::Remove( Node* n1 )
{
	std::list<NodeConnect>::iterator it = n1->m_connectNode.begin();
	for ( ; it != n1->m_connectNode.end(); it++ ) {
		//ڑ̃m[h玩g̏O
		it->node->RemoveConnect(n1);
	}

	//Xg玩MOĉB
	m_node.remove(n1);
	delete n1;
}

//m[hڑ폜
void Dijkstra::RemoveConnect( Node* n1, Node* n2 )
{
	//n1̃m[hn2O
	n1->RemoveConnect(n2);
	n2->RemoveConnect(n1);
}

//Sm[h폜
void Dijkstra::Clear()
{
	std::list<Node*>::iterator it = m_node.begin();
	for ( ; it != m_node.end(); it++ ) {
		delete (*it);
	}
	
	m_node.clear();
	m_start = NULL;
	m_goal  = NULL;
}

//m[hXg̉Ԗڂ 0X^[g Ƃ-1
int Dijkstra::QueryIdx( Node* node )
{
	int ret = 0;
	std::list<Node*>::iterator it = m_node.begin();
	for ( ; it != m_node.end(); it++, ret++ ) {
		if ( *it == node ) {
			return ret;
		}
	}
	return -1;
}

//CfbNXwŃXgm[h擾
Node* Dijkstra::QueryFromIdx(int idx)
{
	int ret = 0;
	std::list<Node*>::iterator it = m_node.begin();
	for ( ; it != m_node.end(); it++, ret++ ) {
		if ( ret == idx ) {
			return (*it);
		}
	}
	return NULL;
}

//Q_Ԃ̃[g쐬
void Dijkstra::ConnectNode( Node* n1, Node* n2 )
{
	//o^ς݂łȂ
	std::list<NodeConnect>::iterator it = n1->m_connectNode.begin();
	for ( ; it != n1->m_connectNode.end(); it++ ) {

		if ( it->node == n2 ) {
			return;
		}
	}
	
	//RXg=vZ
	double cost = QueryCost(n1, n2);

	//ڑ
	n1->m_connectNode.push_back( NodeConnect(n2,(int)cost) );
	n2->m_connectNode.push_back( NodeConnect(n1,(int)cost) );
}

//NbNʒũm[h擾
Node* Dijkstra::Query(int x, int y)
{
	std::list<Node*>::iterator it = m_node.begin();
	for ( ; it != m_node.end(); it++ ) {
		
		if ( (*it)->x == x && (*it)->y == y ) {
			return (*it);	
		}
	
	}
	return NULL;
}

// ԍm[h擾
Node* Dijkstra::Query(int iCastleID)
{
	std::list<Node*>::iterator it = m_node.begin();
	for ( ; it != m_node.end(); it++ ) {
		
		if ( (*it)->GetCastleID() == iCastleID ) {
			return (*it);	
		}
	
	}
	return NULL;
}

//NbNʒu̐ɏm[hg擾
bool Dijkstra::QueryLineOnNodePair( NodePair* rslt, int x, int y )
{
	int search_leng = Circle_r * 1;

	std::list<Node*>::iterator it = m_node.begin();
	Node *node, *sideNode;
	for ( ; it != m_node.end(); it++ ) {
		node = *it;
		
		std::list<NodeConnect>::iterator it_cnct = (*it)->m_connectNode.begin();
		for ( ; it_cnct != (*it)->m_connectNode.end(); it_cnct++ ) {
			
			sideNode = it_cnct->node;
		
			//_Ɛ̌`FbN
			if ( IntersectPointLine( x, y, node->x, node->y, sideNode->x, sideNode->y, search_leng ) ) {
				rslt->n1 = node;
				rslt->n2 = sideNode;
				return true;
			}
		}
	}

	return false;
}

//em[hZbg
void Dijkstra::ResetNode() {
	
	std::list<Node*>::iterator it = m_node.begin();
	for ( ; it != m_node.end(); it++ ) {
		(*it)->Reset();
	}
}

//[gJn
void Dijkstra::SearchRoot() {

	//Sm[ȟvZʂZbg
	ResetNode();

	if ( !m_start || !m_goal ) { return; }

	std::list<Node*> work1;
	std::list<Node*> work2;

	std::list<Node*>* currLevel = &work1;
	std::list<Node*>* nextLevel = &work2;
	std::list<Node*>* for_swap;//currLevelnextLevelւɎg

    //S[vZ܂B
	m_goal->cost = 0;//S[ɃRXg0ZbgČvZς݂Ƃ
	currLevel->push_back(m_goal);//Kw̃m[hZbg(S[m[hZbg)

	std::list<Node*>::iterator it;
	int nodeCost;
	while ( currLevel->size() ) {

		for( it = currLevel->begin(); it != currLevel->end(); it++ ) {
			
			//̃m[hKw֐i߂ȂAnextxɊi[
			std::list<NodeConnect>::iterator it_cnct = (*it)->m_connectNode.begin();
			for( ; it_cnct != (*it)->m_connectNode.end(); it_cnct++ ) {
				
				nodeCost = (*it)->cost + it_cnct->cost;

				if ( ( it_cnct->node->cost == -1 ) || ( nodeCost < it_cnct->node->cost ) ) {
					//Tm[h邢͍ŒZ[gXVłꍇ

					//oHRXgƃS[֌߂̃m[hZbg
					it_cnct->node->cost = nodeCost;
					it_cnct->node->toGoal = (*it);

				} else {
					continue;
				}
				
				//ɌKw̃Xgɓo^
				nextLevel->push_back( it_cnct->node );
			}
		
		}
		
		//XgւĎ̊Kw
		for_swap = currLevel;
		currLevel = nextLevel;
		nextLevel = for_swap;
		nextLevel->clear();//NA
	}
		
}